.tags {
  display: flex;
  flex-direction: row;
  gap: 8px;
  margin-top: 12px;
}

.tag {
  cursor: pointer;
  font-size: 12px;
  color: var(--color-text-light);
  background-color: var(--color-background-lighter);
  border-radius: 6px;
  padding: 4px 8px;
}

.authorTag {
  color: var(--color-text-cyan);
  background-color: var(--color-background-cyan);
}
