[project]
name = "{{ project_name }}"
requires-python = ">=3.9,<3.13"
version = "0.1.0"
{{ dependencies }}

[project.entry-points]
"dagster_dg.library" = { {{ project_name }} = "{{ project_name }}.lib"}

[dependency-groups]
{{ dev_dependencies }}

[build-system]
requires = ["setuptools"]
build-backend = "setuptools.build_meta"

[tool.dg]
directory_type = "project"

[tool.dg.project]
root_module = "{{ project_name }}"

[tool.setuptools.packages.find]
exclude=["{{ project_name }}_tests"]

{{ uv_sources }}
