from dagster import Definitions
from dagster_components import (
    Component,
    ComponentLoadContext,
    Resolvable,
)

class {{ name }}(Component, Resolvable):
    """COMPONENT SUMMARY HERE.

    COMPONENT DESCRIPTION HERE.
    """

    def __init__(
        self,
        # added arguments here will define yaml schema via Resolvable
    ):
        pass

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        # Add definition construction logic here.
        return Definitions()
